/*
 * ConsoleForm.cs
 * $Id: ConsoleForm.cs,v 1.3 2003/11/24 10:09:35 Mayuki Sawatari Exp $
 *
 * Copyright (c) 2003 Mayuki Sawatari, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE MISUZILLA.ORG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Collections.Specialized;
using System.Runtime.InteropServices;

using Misuzilla.Net.IRC.Ircv;

namespace
Misuzilla.Net.IRC.Ircv
{
	internal class ConsoleForm : Form
	{
		public delegate void MessageSendEventHandler(String channel, String msg);
		public event MessageSendEventHandler MessageSend;

		// 手抜き
		public Keys ChannelUpDownModifierKey = Keys.Shift;
		public String[] NumAccessChannels = new String[10];

		private StringCollectionCaseInsensitive _channels = new StringCollectionCaseInsensitive();
		private Int32 _curChannelIndex = -1;
		private TextBox _textBox = null;

		//
		public
		ConsoleForm()
		{
			InitializeComponents();
		}
		
		private void
		InitializeComponents()
		{
			// form
			//this.BackColor = Color.White;
			this.ShowInTaskbar = false;
/**/		this.MinimizeBox = false;
			this.MaximizeBox = false;
			this.ControlBox = false;
			this.TopMost = true;
			//this.FormBorderStyle = FormBorderStyle.FixedToolWindow;
			this.FormBorderStyle = FormBorderStyle.None;
			this.Text = "";
			this.Font = new Font("Tahoma", 8);
			
			_textBox = new TextBox();
			_textBox.Anchor = AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Bottom | AnchorStyles.Right;
			_textBox.AutoSize = false;
			_textBox.BorderStyle = BorderStyle.None;
			_textBox.KeyPress += new KeyPressEventHandler(TextBoxKeyPressed);
			_textBox.KeyDown += new KeyEventHandler(TextBoxKeyDown);

			this.Height = _textBox.PreferredHeight + 4 + this.Font.Height;
			_textBox.Bounds = new Rectangle(1, this.Font.Height+2, this.ClientSize.Width - 2, this.ClientSize.Height - (this.Font.Height + 4));
			_textBox.Update();
			this.Controls.Add(_textBox);
			
			this.ContextMenu = new ContextMenu();
			this.ContextMenu.MenuItems.Add(new MenuItem("最小化(&N)", new EventHandler(ContextMenuMinimizeClicked)));
			this.ContextMenu.MenuItems.Add(new MenuItem("-"));
			this.ContextMenu.MenuItems.Add(new MenuItem("キャンセル(&C)"));
		}

		public void
		AddChannel(String channel)
		{
			if (!_channels.Contains(channel)) {
				_channels.Add(channel);
				SetChannels();
			}
		}
		public void
		RemoveChannel(String channel)
		{
			if (!_channels.Contains(channel)) {
				_channels.Remove(channel);
				SetChannels();
			}
		}

		private void
		SetChannels()
		{
			if (!_channels.Contains(this.Text)) {
				this.Text = " ";
				_curChannelIndex = -1;
				foreach (String newchannel in _channels) {
					this.Text = newchannel;
					_curChannelIndex = 0;
					break;
				}
			}
			
			this.ContextMenu.MenuItems.Clear();
			StringCollection scChannelAccesskey = new StringCollection();
			scChannelAccesskey.AddRange(NumAccessChannels);
			foreach (String channel in _channels) {
				Int32 index = scChannelAccesskey.IndexOf(channel);
				System.Text.StringBuilder sb = new System.Text.StringBuilder(100);
				sb.Append(channel);
				if (index > -1) {
					sb.Append("(&");
					sb.Append(index.ToString());
					sb.Append(")");
				}
				this.ContextMenu.MenuItems.Add(new MenuItem(sb.ToString(), new EventHandler(ContextMenuClicked)));
			}
			
			this.ContextMenu.MenuItems.Add(new MenuItem("-"));
			this.ContextMenu.MenuItems.Add(new MenuItem("非表示(&N)", new EventHandler(ContextMenuMinimizeClicked)));
			this.ContextMenu.MenuItems.Add(new MenuItem("-"));
			this.ContextMenu.MenuItems.Add(new MenuItem("キャンセル(&C)"));
		}
		private void
		ContextMenuMinimizeClicked(Object sender, EventArgs e)
		{
			this.Visible = false;
		}
		private void
		ContextMenuClicked(Object sender, EventArgs e)
		{
			this.Text = _channels[((MenuItem)sender).Index];
		}
		
		private void
		TextBoxKeyDown(Object sender, KeyEventArgs e)
		{
			if (_curChannelIndex < 0) return;
			
			if (e.Control && (e.KeyCode == Keys.Space)) {
				// Ctrl + Space でメニュー
				this.ContextMenu.Show(this, new Point(0, 0));
				e.Handled = true;
			}
			if (e.Control) {
				// Ctrl + [0-9] でチャンネル切り替え
				Int32 index = 0;
				if (e.KeyValue >= (Int32)Keys.D0 && e.KeyValue <= (Int32)Keys.D9) {
					index = e.KeyValue - (Int32)Keys.D0;
				} else if (e.KeyValue >= (Int32)Keys.NumPad0 && e.KeyValue <= (Int32)Keys.NumPad9) {
					index = e.KeyValue - (Int32)Keys.NumPad0;
				}
				if (_channels.Contains(NumAccessChannels[index])) {
					this.Text = NumAccessChannels[index];
				}
				e.Handled = true;
			}
			if (ChannelUpDownModifierKey != e.Modifiers) return;

			if ((e.KeyCode == Keys.Up || e.KeyCode == Keys.Down)) {
				if (_curChannelIndex == 0 && e.KeyCode == Keys.Up) {
					_curChannelIndex = _channels.Count - 1;
				} else if (_curChannelIndex == _channels.Count - 1 && e.KeyCode == Keys.Down){
					_curChannelIndex = 0;
				} else if (_curChannelIndex <= _channels.Count - 1) {
					if (e.KeyCode == Keys.Down)
						_curChannelIndex++;
					else
						_curChannelIndex--;
				}
				//Console.WriteLine("{0} : {1}/{2} [{3}]", e.KeyCode, _curChannelIndex, _channels.Count - 1, _channels[_curChannelIndex]);
				this.Text = _channels[_curChannelIndex];
				e.Handled = true;
			}
		}
		private void
		TextBoxKeyPressed(Object sender, KeyPressEventArgs e)
		{
			if (e.KeyChar == (Char)13) {
				// enter でメッセージ送信 delegate
				if (_textBox.Text == "") return;
				if (this.Text.Trim() == "") return;

				if (this.MessageSend != null) {
					this.MessageSend(this.Text, _textBox.Text);
				}
				_textBox.Text = "";
				
				e.Handled = true;
			}
		}
		
		protected override void
		OnMouseDown(MouseEventArgs e)
		{
			base.OnMouseDown(e);
			// フォームドラッグで移動。
			if (e.Button == MouseButtons.Left) {
				Double op = this.Opacity;
				this.Opacity = 0.55;
				ReleaseCapture();
				// Int32 WM_NCLBUTTONDOWN = 0xa1;
				// Int32 HTCAPTION = 0x02;
				SendMessage(this.Handle, 0xa1, 0x02, IntPtr.Zero);
				this.Opacity = op;
			}
		}
		[DllImport("user32.dll")]
		private static extern Int32
		ReleaseCapture();
		
		[DllImport("user32.dll")]
		private static extern Int32
		SendMessage(IntPtr hWnd, Int32 wMsg, Int32 wParam, IntPtr lParam);

		protected override void
		OnActivated(EventArgs e)
		{
			base.OnActivated(e);
			this.Focus();
			this.Refresh();
			_textBox.Focus();
		}
		protected override void
		OnDeactivate(EventArgs e)
		{
			base.OnDeactivate(e);
			this.Refresh();
		}
		protected override void
		OnPaint(PaintEventArgs e)
		{
			base.OnPaint(e);
			
			Color borderColor, fontColor;
			if (Form.ActiveForm == this) {
				borderColor = Color.FromKnownColor(KnownColor.ActiveCaption);
				fontColor = Color.FromKnownColor(KnownColor.ActiveCaptionText);
			} else {
				borderColor = Color.FromKnownColor(KnownColor.InactiveCaption);
				fontColor = Color.FromKnownColor(KnownColor.InactiveCaptionText);
			}
			
			e.Graphics.DrawRectangle(new Pen(borderColor), new Rectangle(0, 0, this.ClientSize.Width - 1, this.ClientSize.Height - 1));
			e.Graphics.FillRectangle(new SolidBrush(borderColor), new Rectangle(2, 2, this.ClientSize.Width - 4, this.Font.Height+2));
			e.Graphics.DrawString(this.Text, this.Font, new SolidBrush(fontColor), 2F, 2F);
		}
		protected override void
		OnTextChanged(EventArgs e)
		{
			base.OnTextChanged(e);
			this.Refresh();
		}
	}
	class TextBoxWithBackText : TextBox
	{
		public String BackText = "";
		public  
		TextBoxWithBackText()
		{
			this.ResizeRedraw = true;
		}
		protected override void
		OnTextChanged(EventArgs e)
		{
			base.OnTextChanged(e);
			PaintBackgroundText();
		}
		private void
		PaintBackgroundText()
		{
			Console.WriteLine(BackText);
			Graphics g = this.CreateGraphics();//Graphics.FromHwnd(this.Handle);
			g.DrawString(BackText, this.Font, new SolidBrush(Color.Gray), this.Width - g.MeasureString(BackText, this.Font).Width, 0F);
			g.Dispose();
			g = null;
		}
	}
	
	internal class
	StringCollectionCaseInsensitive : StringCollection
	{
		public new Boolean
		Contains(String s)
		{
			for (Int32 i = 0; i < this.Count; i++) {
				if (String.Compare(s, this[i], true) == 0) return true;
			}
			return false;
		}
	}
}
